tfmodel_path = "./tf_model"
PRECISION = "INT8" # Options are "FP32", "FP16", or "INT8"

# --------------------- load test img for performance test ------------------------
import numpy as np
import time
import cv2

BATCH_SIZE = 1

url='./n02099601_3004.jpg'
img = cv2.imread(url)
img = cv2.resize(img,(32, 32))
dummy_input_batch = 255*np.array(np.repeat(np.expand_dims(np.array(img, dtype=np.float32), axis=0), BATCH_SIZE, axis=0), dtype=np.float32)
# ---------------------------- end of loading test img -----------------------------

import tensorflow as tf
from tensorflow.keras import datasets
from tensorflow.python.compiler.tensorrt import trt_convert as tf_trt
from tensorflow.python.saved_model import tag_constants
import os

# -------------------- load test dateset --------------------------------------
# 1. 归一化函数实现；cifar10 均值和方差，自己计算的。
img_mean = tf.constant([0.50736203482434500, 0.4866895632914611, 0.4410885713465068])
img_std  = tf.constant([0.26748815488001604, 0.2565930997269337, 0.2763085095510783])
def normalize(x, mean=img_mean, std=img_std):
    x = (x - mean)/std
    return x

# 2. 数据预处理，仅仅是类型的转换。    [-1~1]
def preprocess(x, y):
    x = tf.pad(x, [[4, 4], [4, 4], [0, 0]])    # 上下填充4个0，左右填充4个0，变为[40, 40, 3]
    x = tf.image.random_crop(x, [32, 32, 3])
    x = tf.image.random_flip_left_right(x)
    # x: [0,255]=> 0~1   其次：normalizaion
    x = tf.cast(x, dtype=tf.float32) / 255.
    # 0~1 => D(0,1) 调用函数；
    x = normalize(x)
    y = tf.cast(y, dtype=tf.int32)
    return x, y

# 数据集的加载
(x, y), (x_test, y_test) = datasets.cifar10.load_data()
y = tf.squeeze(y)            # 或者tf.squeeze(y, axis=1)把1维度的squeeze掉。
y_test = tf.squeeze(y_test)  # 或者tf.squeeze(y, axis=1)把1维度的squeeze掉。

# 测试集Dataset
test_db = tf.data.Dataset.from_tensor_slices((x_test, y_test))
test_db = test_db.map(preprocess).batch(1)
# -------------------- end of loading dateset --------------------------------------

if PRECISION == "INT8":
    def calibration_input_fn():
        for i,(x, y) in enumerate(test_db):
            yield (x, )
            if i ==5:
                break

precision_dict = {
    "FP32": tf_trt.TrtPrecisionMode.FP32,
    "FP16": tf_trt.TrtPrecisionMode.FP16,
    "INT8": tf_trt.TrtPrecisionMode.INT8,
}
trt_precision = precision_dict[PRECISION]

# 参考官方文档的保存及读取trt模型方法：
# https://docs.nvidia.com/deeplearning/frameworks/tf-trt-user-guide/index.html#worflow-with-savedmodel
output_saved_model_dir = tfmodel_path + '_' + PRECISION

if not os.path.exists(output_saved_model_dir):
    # 保存trt加速后的模型
    conversion_params = tf_trt.DEFAULT_TRT_CONVERSION_PARAMS
    conversion_params = conversion_params._replace(precision_mode=trt_precision,
                                                    use_calibration= PRECISION == "INT8")
    converter = tf_trt.TrtGraphConverterV2(
        input_saved_model_dir=tfmodel_path,
        conversion_params=conversion_params)
    if PRECISION == "INT8":
        converter.convert(calibration_input_fn=calibration_input_fn)
    else:
        converter.convert()
    converter.save(output_saved_model_dir)
# 读取trt加速后的模型
saved_model_loaded = tf.saved_model.load(
    output_saved_model_dir, tags=[tag_constants.SERVING])
# signature_keys = list(saved_model_loaded.signatures.keys())
infer = saved_model_loaded.signatures['serving_default']

model_trt = infer

# ------------------- test acc ------------------------
total_num = 0
total_correct = 0
for i, (x, y) in enumerate(test_db):
    output = model_trt(x)['output_1'].numpy()
    prob = tf.nn.softmax(output, axis=1)
    pred = tf.argmax(prob, axis=1)
    pred = tf.cast(pred, dtype=tf.int32) # pred类型为int64,需要转换一下。
    # 拿到预测值pred和真实值比较。
    correct = tf.cast(tf.equal(pred, y), dtype=tf.int32)
    correct = tf.reduce_sum(correct)
    total_num += x.shape[0]
    total_correct += int(correct)  # 转换为numpy数据

acc = total_correct / total_num
print('test_acc:', acc)
# ------------------- end of acc test ------------------------

# -------------------------- test performance -------------------------
dummy_input_batch = tf.constant(dummy_input_batch.astype('float32'))

time_list = []
for i in range(100):
    time1 = time.time()
    trt_output = model_trt(dummy_input_batch)['output_1']
    time2 = time.time()
    time_list.append(time2-time1)

# print("trt_output is ", trt_output)
print("input shape is ", dummy_input_batch.shape)

# print("100 time is ", time_list)

arr_mean = np.mean(time_list) # 求均值
arr_var = np.var(time_list) # 求方差

print("所有数据均值：", arr_mean)
print("所有数据方差：", arr_var)

time_list = time_list[50:]

arr_mean = np.mean(time_list) # 求均值
arr_var = np.var(time_list) # 求方差

print("后50次均值：", arr_mean)
print("后50次方差：", arr_var)
# ------------------- end of performance test -------------------------
